(* ::Package:: *)

(* ::Subtitle:: *)
(*Mathematica ROOT Importer*)


(* ::Section::Closed:: *)
(*Begin Converter Context*)


Begin["ROOTImport`"]


(* ::Section::Closed:: *)
(*Debug Switches*)


$Debug = True;
$Error = True;


DPrint[x___]:=If[$Debug, Print[x]];
EPrint[x___]:=If[$Error, Print[x]];


(* ::Section:: *)
(*Import*)


(* ::Subsection::Closed:: *)
(*Documentation of Functions from the MathLink Portion*)


(*******************************************************************************
The functions that come from MathLink all start with "import*" and "export*".
These functions call the C++ functions "iimport*" or "eexport*".
********************************************************************************)


(*******************************************************************************
The functions coming from MathLink are:
  ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ __
  importKey[filename_String]
     input: filename with absolute path (ex: "/home/ken/Documents/data.root")
     output: a list quartets of the form
             { {Name1, Title1, Class1}, {Name2, Title2, Class2}, ... }
             as given by TFile::GetListOfKeys(), organized by
             {TKey::GetName(), TKey::GetTitle(), TKey::GetClassName(), TKey::GetSeekKey() }
  ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ __
  importTreeMetadata[filename_String, treename_String]
     input: filename with absolute path (ex: "/home/ken/Documents/data.root")
            treename: name of TTree object, given as a string
     output: a list of strings given by TTree:Print ().
     comment: this function is wrapped by the Mathematica function 
                ImportTreeMetadata[treename][filename]
              which calls the MathLink function and processes the list of strings to give
              a list of quarters of the form
              { {BranchName1, BranchTitle1, Type1, NEntries1},
                {BranchName2, BranchTitle2, Type2, NEntries2},
                 ... }
  ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ __
  importBranch[filename_String, tree_String, br_String, type_Integer, beg_Integer, end_Integer]
     input: filename with absolute path (ex: "/home/ken/Documents/data.root")
            tree: treename, given as a string
            br: branch name, given as a string
            type: the type of the data in the branch, given as an integer with the key looked up via $DataTypeDictionary
     output: a list of data of the corresponding type
  ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ __
   importTH1F[filename_String, hist_String]
     input: filename with absolute path (ex: "/home/ken/Documents/data.root")
            hist: name of TH1F object
     output: a list quartets:
             { {bin_lower _edge _ 1, bin_width _ 1, bin_content _ 1, bin_error _ 1},
               {bin_lower _edge _ 2, bin_width _ 2, bin_content _ 2, bin_error _ 2},
               ...
             }
  ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ __
   importTH2F[filename_String, hist_String]
     input: filename with absolute path (ex: "/home/ken/Documents/data.root")
            hist: name of TH2F object
     output: a list of items, each with six entries:
          { {x_lower _edge _ 1, x_width _ 1, y_lower _edge _ 1, y_width _ 1, bin_content _ 1, bin_error _ 1},
            {x_lower _edge _ 2, x_width _ 2, y_lower _edge _ 2, y_width _ 2, bin_content _ 2, bin_error _ 2},
            ...
          }
  ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ __
********************************************************************************)


(* ::Subsection::Closed:: *)
(*Set Root Directory and BuildLibraries*)


$LockFile = FileNameJoin[{$UserBaseDirectory,"SystemFiles","Formats","ROOT","CreateLibrary.m"}]


SetDirectoryQ = False;
SetRootDirectory[]:= If[ !SetDirectoryQ,
 SetDirectoryQ = SetDir[];
 If[ FileExistsQ[$LockFile],
     LoadLibraries[],
     BuildLibraries[]; Export[$LockFile, "Libraries generated: "~~DateString[], "Text" ];
 ]
]


(* ::Subsection::Closed:: *)
(*TKey*)


CheckROOT[file_String]:=
Module[
  {strm, res},
  strm = OpenRead[file, BinaryFormat -> True];
  res = BinaryReadList[file, "Character8", 4];
  Close[strm];
  res === {"r","o","o","t"}
]


(*****************************************************************
This function is registered as the importer of the "Keys" element,
so it takes filename as input, and returns ("Keys" -> result) if it succeeds.
*****************************************************************)
ImportKey[file_, opts___]:=
Module[
  {key},
  SetRootDirectory[];
  If[ !CheckROOT[file], EPrint["Not a valid ROOT file"]; Return[$Failed]; ];
  key = importKey[file];
  If[ key === $Failed, $Failed, "Keys" -> key ]
]


(* ::Subsection:: *)
(*TTree Data & Metadata*)


(* ::Subsubsection::Closed:: *)
(*Metadata*)


(*********************************************************************
This function iteratively calls the MathLink function
   importTreeMetadata[filename, tree]
and process the output strings to give a list of triplets
{branch name, branch type, entries}.
********************************************************************)
ImportTreeMetadata[tree_][file_, opts___]:=
Module[
  {keys, out},
  SetRootDirectory[];
  If[ !CheckROOT[file], EPrint["Not a valid ROOT file"]; Return[$Failed]; ];
  keys = importKey[file];
  If[!MemberQ[First/@Select[keys, MatchQ[#[[3]],"TTree"|"TNtuple"]&], tree], EPrint["No such TTree exists"]; Return[$Failed]];
  out = importTreeMetadata[file, tree];
  "TTreeMetadata" -> {tree -> out}
]


(* ::Subsubsection::Closed:: *)
(*Data*)


(* ::Subsubsubsection::Closed:: *)
(*$DataTypeDictionary*)


$DataTypeDictionary = {
"Int_t" -> 1,
"UInt_t" -> 2,
"Float_t" -> 3,
"Double_t" -> 4,
"Short_t" -> 5,
"UShort_t"-> 6,
"Bool_t" -> 7,
"string" -> 8,
"Char_t" -> 10,

"int" -> 1,
"unsigned int" -> 2,
"float" -> 3,
"double" -> 4,
"short" -> 5,
"unsigned short"-> 6,
"bool" -> 7,

"vector<int>" -> 101,
"vector<unsigned int>" -> 102,
"vector<float>" -> 103,
"vector<double>" -> 104,
"vector<short>" -> 105,
"vector<unsigned short>" -> 106,
"vector<bool>" -> 107,
"vector<string>"-> 108,

"vector<vector<int> >" -> 201,
"vector<vector<unsigned int> >" -> 202,
"vector<vector<float> >" -> 203,
"vector<vector<double> >" -> 204,
"vector<vector<short> >" -> 205,
"vector<vector<unsigned short> >" -> 206,
"vector<vector<bool> >" -> 207,
"vector<vector<string> >" -> 208,

"list<int>" -> 301,
"list<unsigned int>" -> 302,
"list<float>" -> 303,
"list<double>" -> 304,
"list<short>" -> 305,
"list<unsigned short>" -> 306,
"list<bool>" -> 307,
"list<string>"-> 308
}


(* ::Subsubsubsection::Closed:: *)
(*Branch Importer*)


(* only known types defined in $DataTypeDictionary gets passed to MathLink *)
importBranch[file_String, tree_String, branch_String, type_String, beg_Integer, end_Integer]:= $Failed;


(*********************************************************************
This function iteratively calls the MathLink function
   importSpecificBranch[filename, tree, br, type]
after performing the following:
(1) verifying that tree name is valid
(2) call ImportTreeMetadata[tree][file] to get the names/types of branches.
********************************************************************)
ImportTreeData[tree_][file_, opts___]:=
Module[
  {temp, out, bounds, beg, end, len},
  SetRootDirectory[];
  If[ !CheckROOT[file], EPrint["Not a valid ROOT file"]; Return[$Failed]; ];
  bounds = "Range" /. Flatten[{opts}];

  If[ MatchQ[bounds, {_Integer, _Integer}],
      {beg, end} = bounds;
      If[ (beg<0)||(end<0), EPrint["Bounds need to be positive."]; Return[$Failed, Module] ];
      If[ beg > end, EPrint["Bounds need to be in increasing order."]; Return[$Failed, Module] ];
  ];

  (* validate such TTree exists *)
  temp = importKey[file];
  temp = First/@Select[temp, MatchQ[#[[3]],"TTree"|"TNtuple"]&]; (* get names of TTree *)
  If[ !MemberQ[temp, tree], 
      EPrint["Allowed TTrees are: ", temp, " and ", tree, " is invalid."]; Return[$Failed]
  ];

  (* fetch branch names and types *)
  temp = importTreeMetadata[file, tree]; (* names of branches *)
  len = Last[First[temp]];
  If[ MatchQ[end, _Integer] && (end > len), EPrint["Inaccessible bound; the branches have only ", len, " entries."]; Return[$Failed, Module] ];
  temp = Map[{#[[1]], #[[3]]/.$DataTypeDictionary}&, temp];

  If[ Select[ Last/@temp, StringQ ] =!={}, EPrint["Data types ", Select[ Last/@temp, StringQ ], " are not supported."]; ];
  If[ MatchQ[ {beg, end}, {_Integer, _Integer}],
      out = Map[ importBranch[file, tree, #[[1]], #[[2]], beg, end]&, temp],
      out = Map[ importBranch[file, tree, #[[1]], #[[2]], 0, 0]&, temp]
  ];
  "TTreeData" -> {tree -> out}
]


(*********************************************************************
This function essentially wraps around the MathLink function
   importSpecificBranchPartial[filename, tree, br, type, beg, end]
after the steps:
(1) verifying that tree and branch names are valid, and
(2) verifying that the bounds are appropriate
(2) finding the appropriate type of data contained in the branch.
********************************************************************)
ImportBranch[tree_, branch_, beg_, end_][file_, opts___]:=
Module[
  {temp, type, bounds, ibeg = beg, iend = end},
  SetRootDirectory[];
  If[ !CheckROOT[file], EPrint["Not a valid ROOT file"]; Return[$Failed]; ];
  bounds = "Range" /. Flatten[{opts}];
  If[ MatchQ[bounds, {_Integer, _Integer}],
      {ibeg, iend} = bounds;
      DPrint["Partial branch import with bounds: ", {ibeg, iend}];
      If[ ibeg > iend, EPrint["Bounds need to be in increasing order."]; Return[$Failed, Module] ];
  ];
  If[ (ibeg<0)||(iend<0), EPrint["Bounds need to be positive."]; Return[$Failed, Module] ];
  temp = importKey[file];

  (* verify for valid TTree name *)
  temp = Select[temp, MatchQ[#[[3]],"TTree"|"TNtuple"]&]; (* get names of TTree *)
  If[ !MemberQ[First/@temp, tree], 
      EPrint["Allowed TTrees are: ", First/@temp, " and ", tree, " is invalid."]; Return[$Failed]
  ];

  (* verify for valid TBranch name *)
  temp = importTreeMetadata[file, tree];
  If[ !MemberQ[First/@temp, branch],
      EPrint["Allowed TBranches are: ", First/@temp, " and ", branch, " is invalid."]; Return[$Failed]
  ];
  temp = Select[temp, (First[#] === branch)&][[1]];
  If[ iend > Last[temp], EPrint["Bound not accessible; the branch ", branch, " has only ", Last[temp], " entires."]; Return[$Failed, Module]; ];

  (* easier to convert to integer so we can use switch in C/C++ *)
  type = temp[[3]] /.$DataTypeDictionary;
  If[ StringQ[type], EPrint["Data type ", type, " is not supported."]; Return[$Failed]; ];

  temp = importBranch[file, tree, branch, type, ibeg, iend];
  If[ (beg===0)&&(end===0),
      Fold[ {Rule[#2, #1]}&, temp, {branch, tree, "TTreeData"} ],
      Fold[ {Rule[#2, #1]}&, temp, {end, beg, branch, tree, "TTreeData"} ]
  ]
]


(* ::Subsection::Closed:: *)
(*TH1F & TH2F Data and Graphics*)


ImportTHData[hist_String, type_String][file_String, opts___]:=
Module[
  {temp, data},
  SetRootDirectory[];
  If[ !CheckROOT[file], EPrint["Not a valid ROOT file"]; Return[$Failed]; ];
  temp = importKey[file];
  temp = First/@Select[temp, #[[3]]===type&]; (* get names of TH1F *)
  If[ !MemberQ[temp, hist], 
      EPrint["Allowed TH1F are: ", temp, " and ", hist, " is invalid."]; Return[$Failed]
  ];

  (* fetch data and return with signature required by import/export framework *)
  Which[
    type === "TH1F",
    data = importTH1F[file, hist];
    "TH1FData" -> {hist -> data}
    ,
    type === "TH2F",
    data = importTH2F[file, hist];
    "TH2FData" -> {hist -> data}
    ,
    True,
    $Failed
  ] (* No semi-colon here!  The output of Which[] is returned by the function. *)
]


ImportTHGraphics[hist_String, type_String][file_String, opts___]:=
Module[
  {temp, data, low, xlist, ylist},
  SetRootDirectory[];
  If[ !CheckROOT[file], EPrint["Not a valid ROOT file"]; Return[$Failed]; ];
  temp = importKey[file];
  temp = First/@Select[temp, #[[3]]===type&]; (* get names of TH1F *)
  If[ !MemberQ[temp, hist], 
      EPrint["Allowed TH1F are: ", temp, " and ", hist, " is invalid."]; Return[$Failed]
  ];

  (* fetch data and return with signature required by import/export framework *)
  Which[
    type === "TH1F",
    data = importTH1F[file, hist];
    low = First/@data;
    low = Append[low, Last[low]+(Last[data][[2]])];
    "TH1FGraphics" -> {hist -> Histogram[{0}, {low}, ((data[[All,3]])&), Sequence@@Flatten[{opts}]] }
    ,
    type === "TH2F",
    data = importTH2F[file, hist];
    xlist=Sort@DeleteDuplicates@data[[All,1]];
    ylist=Sort@DeleteDuplicates@data[[All,3]];
    xlist = Developer`ToPackedArray@Append[xlist, Last[xlist]+Last[data][[2]]];
    ylist = Developer`ToPackedArray@Append[ylist, Last[ylist]+Last[data][[4]]];
    data = (Partition[data[[All,5]], Length[ylist]-1 ]);
    "TH2FGraphics" -> {hist -> Histogram3D[{{0,0}}, {{xlist},{ylist}}, (data)&, Sequence@@Flatten[{opts}]] }
    ,
    True,
    $Failed
  ]
]


(* ::Section::Closed:: *)
(*End Converter Context*)


End[]
